@@ -10,7 +10,7 @@ use pyo3::types::PyTuple;
 
 use crate::bytecode;
 use crate::error::{
-    message_bad_eof, message_from_token, message_incorrect_requirement, QASM2ParseError,
+    message_bad_eof, message_generic, message_incorrect_requirement, Position, QASM2ParseError,
 };
 use crate::lex::{Token, TokenContext, TokenStream, TokenType};
 use crate::parse::{GateSymbol, GlobalSymbol};
@@ -234,12 +234,12 @@ impl<'a> ExprParser<'a> {
     /// Get the next token available in the stack of token streams, popping and removing any
     /// complete streams, except the base case.  Will only return `None` once all streams are
     /// exhausted.
-    fn next_token(&mut self) -> Option<Token> {
+    fn next_token(&mut self) -> PyResult<Option<Token>> {
         let mut pointer = self.tokens.len() - 1;
         while pointer > 1 {
-            let out = self.tokens[pointer].next(self.context);
+            let out = self.tokens[pointer].next(self.context)?;
             if out.is_some() {
-                return out;
+                return Ok(out);
             }
             self.tokens.pop();
             pointer -= 1;
@@ -249,9 +249,9 @@ impl<'a> ExprParser<'a> {
 
     /// Peek the next token in the stack of token streams.  This does not remove any complete
     /// streams yet.  Will only return `None` once all streams are exhausted.
-    fn peek_token(&mut self) -> Option<&Token> {
+    fn peek_token(&mut self) -> PyResult<Option<&Token>> {
         let mut pointer = self.tokens.len() - 1;
-        while pointer > 1 && self.tokens[pointer].peek(self.context).is_none() {
+        while pointer > 1 && self.tokens[pointer].peek(self.context)?.is_none() {
             pointer -= 1;
         }
         self.tokens[pointer].peek(self.context)
@@ -266,12 +266,15 @@ impl<'a> ExprParser<'a> {
     /// [parse::State::expect].  The error variant of the result contains a suitable error message
     /// if the expectation is violated.
     fn expect(&mut self, expected: TokenType, required: &str, cause: &Token) -> PyResult<Token> {
-        let token = match self.next_token() {
+        let token = match self.next_token()? {
             None => {
                 return Err(QASM2ParseError::new_err(message_bad_eof(
-                    self.current_filename(),
+                    Some(&Position::new(
+                        self.current_filename(),
+                        cause.line,
+                        cause.col,
+                    )),
                     required,
-                    cause,
                 )))
             }
             Some(token) => token,
@@ -280,19 +283,19 @@ impl<'a> ExprParser<'a> {
             Ok(token)
         } else {
             Err(QASM2ParseError::new_err(message_incorrect_requirement(
-                self.current_filename(),
                 required,
                 &token,
+                self.current_filename(),
             )))
         }
     }
 
     /// Peek the next token from the stream, and consume and return it only if it has the correct
     /// type.
-    fn accept(&mut self, acceptable: TokenType) -> Option<Token> {
-        match self.peek_token() {
+    fn accept(&mut self, acceptable: TokenType) -> PyResult<Option<Token>> {
+        match self.peek_token()? {
             Some(Token { ttype, .. }) if *ttype == acceptable => self.next_token(),
-            _ => None,
+            _ => Ok(None),
         }
     }
 
@@ -316,10 +319,13 @@ impl<'a> ExprParser<'a> {
     fn apply_infix(&mut self, infix: Op, lhs: Expr, rhs: Expr, op_token: &Token) -> PyResult<Expr> {
         if let (Expr::Constant(val), Op::Divide) = (&rhs, infix) {
             if *val == 0.0 {
-                return Err(QASM2ParseError::new_err(message_from_token(
-                    op_token,
+                return Err(QASM2ParseError::new_err(message_generic(
+                    Some(&Position::new(
+                        self.current_filename(),
+                        op_token.line,
+                        op_token.col,
+                    )),
                     "cannot divide by zero",
-                    self.current_filename(),
                 )));
             }
         };
@@ -358,13 +364,16 @@ impl<'a> ExprParser<'a> {
                     if val > 0.0 {
                         Ok(Expr::Constant(val.ln()))
                     } else {
-                        Err(QASM2ParseError::new_err(message_from_token(
-                            token,
+                        Err(QASM2ParseError::new_err(message_generic(
+                            Some(&Position::new(
+                                self.current_filename(),
+                                token.line,
+                                token.col,
+                            )),
                             &format!(
                                 "failure in constant folding: cannot take ln of non-positive {}",
                                 val
                             ),
-                            self.current_filename(),
                         )))
                     }
                 }
@@ -373,13 +382,16 @@ impl<'a> ExprParser<'a> {
                     if val >= 0.0 {
                         Ok(Expr::Constant(val.sqrt()))
                     } else {
-                        Err(QASM2ParseError::new_err(message_from_token(
-                            token,
+                        Err(QASM2ParseError::new_err(message_generic(
+                            Some(&Position::new(
+                                self.current_filename(),
+                                token.line,
+                                token.col,
+                            )),
                             &format!(
                                 "failure in constant folding: cannot take sqrt of negative {}",
                                 val
                             ),
-                            self.current_filename(),
                         )))
                     }
                 }
@@ -411,23 +423,27 @@ impl<'a> ExprParser<'a> {
                     }),
                 );
                 match callable.call1(py, args) {
-                    Ok(retval) => match retval.extract::<f64>(py) {
-                        Ok(fval) => Ok(Expr::Constant(fval)),
-                        Err(inner) => {
-                            let error = QASM2ParseError::new_err(message_from_token(
-                                token,
+                    Ok(retval) => {
+                        match retval.extract::<f64>(py) {
+                            Ok(fval) => Ok(Expr::Constant(fval)),
+                            Err(inner) => {
+                                let error = QASM2ParseError::new_err(message_generic(
+                                Some(&Position::new(self.current_filename(), token.line, token.col)),
                                 "user-defined function returned non-float during constant folding",
-                                self.current_filename(),
                             ));
-                            error.set_cause(py, Some(inner));
-                            Err(error)
+                                error.set_cause(py, Some(inner));
+                                Err(error)
+                            }
                         }
-                    },
+                    }
                     Err(inner) => {
-                        let error = QASM2ParseError::new_err(message_from_token(
-                            token,
+                        let error = QASM2ParseError::new_err(message_generic(
+                            Some(&Position::new(
+                                self.current_filename(),
+                                token.line,
+                                token.col,
+                            )),
                             "caught exception when constant folding with user-defined function",
-                            self.current_filename(),
                         ));
                         error.set_cause(py, Some(inner));
                         Err(error)
@@ -442,10 +458,13 @@ impl<'a> ExprParser<'a> {
     /// If in `strict` mode, and we have a trailing comma, emit a suitable error message.
     fn check_trailing_comma(&self, comma: Option<&Token>) -> PyResult<()> {
         match (self.strict, comma) {
-            (true, Some(token)) => Err(QASM2ParseError::new_err(message_from_token(
-                token,
+            (true, Some(token)) => Err(QASM2ParseError::new_err(message_generic(
+                Some(&Position::new(
+                    self.current_filename(),
+                    token.line,
+                    token.col,
+                )),
                 "[strict] trailing commas in parameter and qubit lists are forbidden",
-                self.current_filename(),
             ))),
             _ => Ok(()),
         }
@@ -478,10 +497,9 @@ impl<'a> ExprParser<'a> {
                 match self.gate_symbols.get(id) {
                     Some(GateSymbol::Parameter { index }) => Ok(Some(Atom::Parameter(*index))),
                     Some(GateSymbol::Qubit { .. }) => {
-                        Err(QASM2ParseError::new_err(message_from_token(
-                            token,
+                        Err(QASM2ParseError::new_err(message_generic(
+                            Some(&Position::new(self.current_filename(), token.line, token.col)),
                             &format!("'{}' is a gate qubit, not a parameter", id),
-                            self.current_filename(),
                         )))
                     }
                     None => {
@@ -490,14 +508,13 @@ impl<'a> ExprParser<'a> {
                                 Ok(Some(Atom::CustomFunction(callable.clone(), *n_params)))
                             }
                             _ =>  {
-                            Err(QASM2ParseError::new_err(message_from_token(
-                            token,
+                            Err(QASM2ParseError::new_err(message_generic(
+                            Some(&Position::new(self.current_filename(), token.line, token.col)),
                             &format!(
                                 "'{}' is not a parameter or custom instruction defined in this scope",
                                 id,
-                            ),
-                            self.current_filename(),)))
-                        }
+                            ))))
+                            }
                     }
                 }
             }
@@ -509,15 +526,15 @@ impl<'a> ExprParser<'a> {
     /// Peek at the next [Atom] (and backing [Token]) if the next token exists and can be converted
     /// into a valid [Atom].  If it can't, or if we are at the end of the input, the `None` variant
     /// is returned.
-    fn peek_atom(&mut self) -> Option<(Atom, Token)> {
-        if let Some(&token) = self.peek_token() {
+    fn peek_atom(&mut self) -> PyResult<Option<(Atom, Token)>> {
+        if let Some(&token) = self.peek_token()? {
             if let Ok(Some(atom)) = self.try_atom_from_token(&token) {
-                Some((atom, token))
+                Ok(Some((atom, token)))
             } else {
-                None
+                Ok(None)
             }
         } else {
-            None
+            Ok(None)
         }
     }
 
@@ -529,26 +546,29 @@ impl<'a> ExprParser<'a> {
     /// and its parsing would finish when it saw the next `+` binary operation.  For initial entry,
     /// the `power_min` should be zero.
     fn eval_expression(&mut self, power_min: u8, cause: &Token) -> PyResult<Expr> {
-        let token = self.next_token().ok_or_else(|| {
+        let token = self.next_token()?.ok_or_else(|| {
             QASM2ParseError::new_err(message_bad_eof(
-                self.current_filename(),
+                Some(&Position::new(
+                    self.current_filename(),
+                    cause.line,
+                    cause.col,
+                )),
                 if power_min == 0 {
                     "an expression"
                 } else {
                     "a missing operand"
                 },
-                cause,
             ))
         })?;
         let atom = self.try_atom_from_token(&token)?.ok_or_else(|| {
             QASM2ParseError::new_err(message_incorrect_requirement(
-                self.current_filename(),
                 if power_min == 0 {
                     "an expression"
                 } else {
                     "a missing operand"
                 },
                 &token,
+                self.current_filename(),
             ))
         })?;
         // First evaluate the "left-hand side" of a (potential) sequence of binary infix operators.
@@ -565,16 +585,22 @@ impl<'a> ExprParser<'a> {
             }
             Atom::RParen => {
                 if power_min == 0 {
-                    Err(QASM2ParseError::new_err(message_from_token(
-                        &token,
+                    Err(QASM2ParseError::new_err(message_generic(
+                        Some(&Position::new(
+                            self.current_filename(),
+                            token.line,
+                            token.col,
+                        )),
                         "did not find an expected expression",
-                        self.current_filename(),
                     )))
                 } else {
-                    Err(QASM2ParseError::new_err(message_from_token(
-                        &token,
+                    Err(QASM2ParseError::new_err(message_generic(
+                        Some(&Position::new(
+                            self.current_filename(),
+                            token.line,
+                            token.col,
+                        )),
                         "the parenthesis closed, but there was a missing operand",
-                        self.current_filename(),
                     )))
                 }
             }
@@ -582,7 +608,7 @@ impl<'a> ExprParser<'a> {
                 let lparen_token =
                     self.expect(TokenType::LParen, "an opening parenthesis", &token)?;
                 let argument = self.eval_expression(0, &token)?;
-                let comma = self.accept(TokenType::Comma);
+                let comma = self.accept(TokenType::Comma)?;
                 self.check_trailing_comma(comma.as_ref())?;
                 self.expect(TokenType::RParen, "a closing parenthesis", &lparen_token)?;
                 Ok(self.apply_function(func, argument, &token)?)
@@ -597,11 +623,11 @@ impl<'a> ExprParser<'a> {
                     // breaking because there are _no_ parameters, because there's a trailing
                     // comma before the closing parenthesis, or because we didn't see a comma after
                     // an expression so we _need_ a closing parenthesis.
-                    if let Some((Atom::RParen, _)) = self.peek_atom() {
+                    if let Some((Atom::RParen, _)) = self.peek_atom()? {
                         break;
                     }
                     arguments.push(self.eval_expression(0, &token)?);
-                    comma = self.accept(TokenType::Comma);
+                    comma = self.accept(TokenType::Comma)?;
                     if comma.is_none() {
                         break;
                     }
@@ -611,14 +637,17 @@ impl<'a> ExprParser<'a> {
                 if arguments.len() == n_params {
                     Ok(self.apply_custom_function(callable, arguments, &token)?)
                 } else {
-                    Err(QASM2ParseError::new_err(message_from_token(
-                        &token,
+                    Err(QASM2ParseError::new_err(message_generic(
+                        Some(&Position::new(
+                            self.current_filename(),
+                            token.line,
+                            token.col,
+                        )),
                         &format!(
                             "custom function argument-count mismatch: expected {}, saw {}",
                             n_params,
                             arguments.len(),
                         ),
-                        self.current_filename(),
                     )))
                 }
             }
@@ -627,10 +656,13 @@ impl<'a> ExprParser<'a> {
                     let expr = self.eval_expression(power, &token)?;
                     Ok(self.apply_prefix(op, expr)?)
                 }
-                None => Err(QASM2ParseError::new_err(message_from_token(
-                    &token,
+                None => Err(QASM2ParseError::new_err(message_generic(
+                    Some(&Position::new(
+                        self.current_filename(),
+                        token.line,
+                        token.col,
+                    )),
                     &format!("'{}' is not a valid unary operator", op.text()),
-                    self.current_filename(),
                 ))),
             },
             Atom::Const(val) => Ok(Expr::Constant(val)),
@@ -640,12 +672,12 @@ impl<'a> ExprParser<'a> {
         // looking at operators that bind more tightly than the `power_min` passed to this
         // function.  Once they're the same power or less, we have to return, because the calling
         // evaluator needs to bind its operator before we move on to the next infix operator.
-        while let Some((Atom::Op(op), peeked_token)) = self.peek_atom() {
+        while let Some((Atom::Op(op), peeked_token)) = self.peek_atom()? {
             let (power_l, power_r) = binary_power(op);
             if power_l < power_min {
                 break;
             }
-            self.next_token(); // Skip peeked operator.
+            self.next_token()?; // Skip peeked operator.
             let rhs = self.eval_expression(power_r, &peeked_token)?;
             lhs = self.apply_infix(op, lhs, rhs, &peeked_token)?;
         }
